from functools import partial as _partial
from .formatstring import fmtstr

black = _partial(fmtstr, style="black")
red = _partial(fmtstr, style="red")
green = _partial(fmtstr, style="green")
yellow = _partial(fmtstr, style="yellow")
blue = _partial(fmtstr, style="blue")
magenta = _partial(fmtstr, style="magenta")
cyan = _partial(fmtstr, style="cyan")
gray = _partial(fmtstr, style="gray")

on_black = _partial(fmtstr, style="on_black")
on_dark = on_black  # deprecated, old name of on_black
on_red = _partial(fmtstr, style="on_red")
on_green = _partial(fmtstr, style="on_green")
on_yellow = _partial(fmtstr, style="on_yellow")
on_blue = _partial(fmtstr, style="on_blue")
on_magenta = _partial(fmtstr, style="on_magenta")
on_cyan = _partial(fmtstr, style="on_cyan")
on_gray = _partial(fmtstr, style="on_gray")

bold = _partial(fmtstr, style="bold")
dark = _partial(fmtstr, style="dark")
italic = _partial(fmtstr, style="italic")
underline = _partial(fmtstr, style="underline")
blink = _partial(fmtstr, style="blink")
invert = _partial(fmtstr, style="invert")

plain = _partial(fmtstr)



r"""Colored strings that behave mostly like strings

>>> s = fmtstr("Hey there!", 'red')
>>> s
red('Hey there!')
>>> s[4:7]
red('the')
>>> red_on_blue = fmtstr('hello', 'red', 'on_blue')
>>> blue_on_red = fmtstr('there', fg='blue', bg='red')
>>> green = fmtstr('!', 'green')
>>> full = red_on_blue + ' ' + blue_on_red + green
>>> full
on_blue(red('hello'))+' '+on_red(blue('there'))+green('!')
>>> str(full)
'\x1b[31m\x1b[44mhello\x1b[49m\x1b[39m \x1b[34m\x1b[41mthere\x1b[49m\x1b[39m\x1b[32m!\x1b[39m'
>>> fmtstr(', ').join(['a', fmtstr('b'), fmtstr('c', 'blue')])
'a'+', '+'b'+', '+blue('c')
>>> fmtstr(u'hello', u'red', bold=False)
red('hello')
"""

import re
from cwcwidth import wcswidth, wcwidth
from functools import cached_property
from itertools import chain
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
    cast,
    no_type_check,
)
from collections.abc import Callable, Iterable, Iterator, Mapping, MutableMapping

from .escseqparse import parse, remove_ansi
from .termformatconstants import (
    FG_COLORS,
    BG_COLORS,
    STYLES,
    FG_NUMBER_TO_COLOR,
    BG_NUMBER_TO_COLOR,
    RESET_ALL,
    RESET_BG,
    RESET_FG,
    seq,
)

one_arg_xforms: Mapping[str, Callable[[str], str]] = {
    "bold": lambda s: seq(STYLES["bold"]) + s + seq(RESET_ALL),
    "dark": lambda s: seq(STYLES["dark"]) + s + seq(RESET_ALL),
    "italic": lambda s: seq(STYLES["italic"]) + s + seq(RESET_ALL),
    "underline": lambda s: seq(STYLES["underline"]) + s + seq(RESET_ALL),
    "blink": lambda s: seq(STYLES["blink"]) + s + seq(RESET_ALL),
    "invert": lambda s: seq(STYLES["invert"]) + s + seq(RESET_ALL),
}

two_arg_xforms: Mapping[str, Callable[[str, int], str]] = {
    "fg": lambda s, v: f"{seq(v)}{s}{seq(RESET_FG)}",
    "bg": lambda s, v: seq(v) + s + seq(RESET_BG),
}


class FrozenAttributes(dict[str, Union[int, bool]]):
    """Immutable dictionary class for format string attributes"""

    def __setitem__(self, key: str, value: int | bool) -> None:
        raise Exception("Cannot change value.")

    def update(self, *args: Any, **kwds: Any) -> None:
        raise Exception("Cannot change value.")

    def extend(self, dictlike: Mapping[str, int | bool]) -> "FrozenAttributes":
        return FrozenAttributes(chain(self.items(), dictlike.items()))

    def remove(self, *keys: str) -> "FrozenAttributes":
        return FrozenAttributes((k, v) for k, v in self.items() if k not in keys)


def stable_format_dict(d: Mapping) -> str:
    """A sorted, python2/3 stable formatting of a dictionary.

    Does not work for dicts with unicode strings as values."""
    inner = ", ".join(
        "{}: {}".format(
            (
                repr(k)[1:]
                if repr(k).startswith("u'") or repr(k).startswith('u"')
                else repr(k)
            ),
            v,
        )
        for k, v in sorted(d.items())
    )
    return "{%s}" % inner


class Chunk:
    """A string with a single set of formatting attributes

    Subject to change, not part of the API"""

    def __init__(self, string: str, atts: Mapping[str, int | bool] | None = None):
        if not isinstance(string, str):
            raise ValueError("unicode string required, got %r" % string)
        self._s = string
        self._atts = FrozenAttributes(atts if atts else {})

    @property
    def s(self) -> str:
        return self._s

    @property
    def atts(self) -> FrozenAttributes:
        "Attributes, e.g. {'fg': 34, 'bold': True} where 34 is the escape code for ..."
        return self._atts

    def __len__(self) -> int:
        return len(self._s)

    @property
    def width(self) -> int:
        width = wcswidth(self._s)
        if len(self._s) > 0 and width < 1:
            raise ValueError("Can't calculate width of string %r" % self._s)
        return width

    @cached_property
    def color_str(self) -> str:
        "Return an escape-coded string to write to the terminal."
        s = self._s
        for k, v in sorted(self._atts.items()):
            # (self.atts sorted for the sake of always acting the same.)
            if k not in one_arg_xforms and k not in two_arg_xforms:
                # Unsupported SGR code
                continue
            elif v is False:
                continue
            elif k in one_arg_xforms:
                s = one_arg_xforms[k](s)
            else:
                s = two_arg_xforms[k](s, v)
        return s

    def __str__(self) -> str:
        value = self.color_str
        if isinstance(value, bytes):
            return value.decode("utf8", "replace")
        return value

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Chunk):
            return NotImplemented
        return self._s == other._s and self._atts == other._atts

    def __hash__(self) -> int:
        return hash((self._s, self._atts))

    def __repr__(self) -> str:
        return "Chunk({s}{sep}{atts})".format(
            s=repr(self._s),
            sep=", " if self._atts else "",
            atts=stable_format_dict(self._atts) if self._atts else "",
        )

    def repr_part(self) -> str:
        """FmtStr repr is build by concatenating these."""

        def pp_att(att: str) -> str:
            if att == "fg":
                return FG_NUMBER_TO_COLOR[self.atts[att]]
            elif att == "bg":
                return "on_" + BG_NUMBER_TO_COLOR[self.atts[att]]
            else:
                return att

        atts_out = {k: v for (k, v) in self._atts.items() if v}
        return (
            "".join(pp_att(att) + "(" for att in sorted(atts_out))
            + repr(self._s)
            + ")" * len(atts_out)
        )

    def splitter(self) -> "ChunkSplitter":
        """
        Returns a view of this Chunk from which new Chunks can be requested.
        """
        return ChunkSplitter(self)


class ChunkSplitter:
    """
    View of a Chunk for breaking it into smaller Chunks.
    """

    def __init__(self, chunk: Chunk) -> None:
        self.reinit(chunk)

    def reinit(self, chunk: Chunk) -> None:
        """Reuse an existing Splitter instance for speed."""
        # TODO benchmark to prove this is worthwhile
        self.chunk = chunk
        self.internal_offset = 0  # index into chunk.s
        self.internal_width = 0  # width of chunks.s[:self.internal_offset]
        divides = [0]
        for c in self.chunk.s:
            divides.append(divides[-1] + wcwidth(c))
        self.divides = divides

    def request(self, max_width: int) -> tuple[int, Chunk] | None:
        """Requests a sub-chunk of max_width or shorter. Returns None if no chunks left."""
        if max_width < 1:
            raise ValueError("requires positive integer max_width")

        s = self.chunk.s
        length = len(s)

        if self.internal_offset == len(s):
            return None

        width = 0
        start_offset = i = self.internal_offset
        replacement_char = " "

        while True:
            w = wcswidth(s[i], None)

            # If adding a character puts us over the requested width, return what we've got so far
            if width + w > max_width:
                self.internal_offset = i  # does not include ith character
                self.internal_width += width

                # if not adding it us makes us short, this must have been a double-width character
                if width < max_width:
                    assert (
                        width + 1 == max_width
                    ), "unicode character width of more than 2!?!"
                    assert w == 2, "unicode character of width other than 2?"
                    return (
                        width + 1,
                        Chunk(
                            s[start_offset : self.internal_offset] + replacement_char,
                            atts=self.chunk.atts,
                        ),
                    )
                return (
                    width,
                    Chunk(s[start_offset : self.internal_offset], atts=self.chunk.atts),
                )
            # otherwise add this width
            width += w

            # If one more char would put us over, return whatever we've got
            if i + 1 == length:
                self.internal_offset = (
                    i + 1
                )  # beware the fencepost, i is an index not an offset
                self.internal_width += width
                return (
                    width,
                    Chunk(s[start_offset : self.internal_offset], atts=self.chunk.atts),
                )
            # otherwise attempt to add the next character
            i += 1


class FmtStr:
    """A string whose substrings carry attributes."""

    def __init__(self, *components: Chunk) -> None:
        # These assertions below could be useful for debugging, but slow things down considerably
        # assert all([len(x) > 0 for x in components])
        # self.chunks = [x for x in components if len(x) > 0]
        self.chunks = list(components)

        # caching these leads to a significant speedup
        self._unicode: str | None = None
        self._len: int | None = None
        self._s: str | None = None
        self._width: int | None = None

    @staticmethod
    def from_str(s: str) -> "FmtStr":
        r"""
        Return a FmtStr representing input.

        The str() of a FmtStr is guaranteed to produced the same FmtStr.
        Other input with escape sequences may not be preserved.

        >>> fmtstr("|"+fmtstr("hey", fg='red', bg='blue')+"|")
        '|'+on_blue(red('hey'))+'|'
        >>> fmtstr('|\x1b[31m\x1b[44mhey\x1b[49m\x1b[39m|')
        '|'+on_blue(red('hey'))+'|'
        """

        if "\x1b[" in s:
            try:
                tokens_and_strings = parse(s)
            except ValueError:
                return FmtStr(Chunk(remove_ansi(s)))
            else:
                chunks = []
                cur_fmt = {}
                for x in tokens_and_strings:
                    if isinstance(x, dict):
                        cur_fmt.update(x)
                    elif isinstance(x, str):
                        atts = parse_args(
                            (), {k: v for k, v in cur_fmt.items() if v is not None}
                        )
                        chunks.append(Chunk(x, atts=atts))
                    else:
                        raise TypeError(f"Expected dict or str, not {type(x)}")
                return FmtStr(*chunks)
        else:
            return FmtStr(Chunk(s))

    def copy_with_new_str(self, new_str: str) -> "FmtStr":
        """Copies the current FmtStr's attributes while changing its string."""
        # What to do when there are multiple Chunks with conflicting atts?
        old_atts = {
            att: value for bfs in self.chunks for (att, value) in bfs.atts.items()
        }
        return FmtStr(Chunk(new_str, old_atts))

    def setitem(self, startindex: int, fs: Union[str, "FmtStr"]) -> "FmtStr":
        """Shim for easily converting old __setitem__ calls"""
        return self.setslice_with_length(startindex, startindex + 1, fs, len(self))

    def setslice_with_length(
        self, startindex: int, endindex: int, fs: Union[str, "FmtStr"], length: int
    ) -> "FmtStr":
        """Shim for easily converting old __setitem__ calls"""
        if len(self) < startindex:
            fs = " " * (startindex - len(self)) + fs
        if len(self) > endindex:
            fs = fs + " " * (endindex - startindex - len(fs))
            assert len(fs) == endindex - startindex, (len(fs), startindex, endindex)
        result = self.splice(fs, startindex, endindex)
        if len(result) > length:
            raise ValueError(
                "Your change is resulting in a longer fmtstr than the original length and this is not supported."
            )
        return result

    def splice(
        self, new_str: Union[str, "FmtStr"], start: int, end: int | None = None
    ) -> "FmtStr":
        """Returns a new FmtStr with the input string spliced into the
        the original FmtStr at start and end.
        If end is provided, new_str will replace the substring self.s[start:end-1].
        """
        if len(new_str) == 0:
            return self
        new_fs = new_str if isinstance(new_str, FmtStr) else fmtstr(new_str)
        assert len(new_fs.chunks) > 0, (new_fs.chunks, new_fs)
        new_components = []
        inserted = False
        if end is None:
            end = start
        tail = None

        for bfs, bfs_start, bfs_end in zip(
            self.chunks, self.divides[:-1], self.divides[1:]
        ):
            if end == bfs_start == 0:
                new_components.extend(new_fs.chunks)
                new_components.append(bfs)
                inserted = True

            elif bfs_start <= start < bfs_end:
                divide = start - bfs_start
                head = Chunk(bfs.s[:divide], atts=bfs.atts)
                tail = Chunk(bfs.s[end - bfs_start :], atts=bfs.atts)
                new_components.extend([head] + new_fs.chunks)
                inserted = True

                if bfs_start < end < bfs_end:
                    tail = Chunk(bfs.s[end - bfs_start :], atts=bfs.atts)
                    new_components.append(tail)

            elif bfs_start < end < bfs_end:
                divide = start - bfs_start
                tail = Chunk(bfs.s[end - bfs_start :], atts=bfs.atts)
                new_components.append(tail)

            elif bfs_start >= end or bfs_end <= start:
                new_components.append(bfs)

        if not inserted:
            new_components.extend(new_fs.chunks)
            inserted = True

        return FmtStr(*(s for s in new_components if s.s))

    def append(self, string: Union[str, "FmtStr"]) -> "FmtStr":
        return self.splice(string, len(self.s))

    def copy_with_new_atts(self, **attributes: bool | int) -> "FmtStr":
        """Returns a new FmtStr with the same content but new formatting"""

        return FmtStr(
            *(Chunk(bfs.s, bfs.atts.extend(attributes)) for bfs in self.chunks)
        )

    def join(self, iterable: Iterable[Union[str, "FmtStr"]]) -> "FmtStr":
        """Joins an iterable yielding strings or FmtStrs with self as separator"""
        before: list[Chunk] = []
        chunks: list[Chunk] = []
        for s in iterable:
            chunks.extend(before)
            before = self.chunks
            if isinstance(s, FmtStr):
                chunks.extend(s.chunks)
            elif isinstance(s, (bytes, str)):
                chunks.extend(fmtstr(s).chunks)  # TODO just make a chunk directly
            else:
                raise TypeError("expected str or FmtStr, %r found" % type(s))
        return FmtStr(*chunks)

    # TODO make this split work like str.split
    def split(
        self,
        sep: str | None = None,
        maxsplit: int | None = None,
        regex: bool = False,
    ) -> list["FmtStr"]:
        """Split based on separator, optionally using a regex.

        Capture groups are ignored in regex, the whole pattern is matched
        and used to split the original FmtStr."""
        if maxsplit is not None:
            raise NotImplementedError("no maxsplit yet")
        s = self.s
        if sep is None:
            sep = r"\s+"
        elif not regex:
            sep = re.escape(sep)
        matches = list(re.finditer(sep, s))
        return [
            self[start:end]
            for start, end in zip(
                chain((0,), (m.end() for m in matches)),
                chain((m.start() for m in matches), (len(s),)),
            )
        ]

    def splitlines(self, keepends: bool = False) -> list["FmtStr"]:
        """Return a list of lines, split on newline characters,
        include line boundaries, if keepends is true."""
        lines = self.split("\n")
        return (
            [line + "\n" for line in lines]
            if keepends
            else (lines if lines[-1] else lines[:-1])
        )

    # proxying to the string via __getattr__ is insufficient
    # because we shouldn't drop foreground or formatting info
    def ljust(self, width: int, fillchar: str | None = None) -> "FmtStr":
        """S.ljust(width[, fillchar]) -> string

        If a fillchar is provided, less formatting information will be preserved
        """
        if fillchar is not None:
            return fmtstr(self.s.ljust(width, fillchar), **self.shared_atts)
        to_add = " " * (width - len(self.s))
        shared = self.shared_atts
        if "bg" in shared:
            return self + fmtstr(to_add, bg=shared["bg"]) if to_add else self
        else:
            uniform = self.new_with_atts_removed("bg")
            return uniform + fmtstr(to_add, **self.shared_atts) if to_add else uniform

    def rjust(self, width: int, fillchar: str | None = None) -> "FmtStr":
        """S.rjust(width[, fillchar]) -> string

        If a fillchar is provided, less formatting information will be preserved
        """
        if fillchar is not None:
            return fmtstr(self.s.rjust(width, fillchar), **self.shared_atts)
        to_add = " " * (width - len(self.s))
        shared = self.shared_atts
        if "bg" in shared:
            return fmtstr(to_add, bg=shared["bg"]) + self if to_add else self
        else:
            uniform = self.new_with_atts_removed("bg")
            return fmtstr(to_add, **self.shared_atts) + uniform if to_add else uniform

    def __str__(self) -> str:
        if self._unicode is not None:
            return self._unicode
        self._unicode = "".join(str(fs) for fs in self.chunks)
        return self._unicode

    def __len__(self) -> int:
        if self._len is not None:
            return self._len
        value = sum(len(fs) for fs in self.chunks)
        self._len = value
        return value

    @property
    def width(self) -> int:
        """The number of columns it would take to display this string."""
        if self._width is not None:
            return self._width
        value = sum(fs.width for fs in self.chunks)
        self._width = value
        return value

    def width_at_offset(self, n: int) -> int:
        """Returns the horizontal position of character n of the string"""
        # TODO make more efficient?
        width = wcswidth(self.s, n)
        assert width != -1
        return width

    def __repr__(self) -> str:
        return "+".join(fs.repr_part() for fs in self.chunks)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, (str, bytes, FmtStr)):
            return str(self) == str(other)
        return NotImplemented

    def __hash__(self) -> int:
        return hash(str(self))

    def __add__(self, other: Union["FmtStr", str]) -> "FmtStr":
        if isinstance(other, FmtStr):
            return FmtStr(*(self.chunks + other.chunks))
        elif isinstance(other, (bytes, str)):
            return FmtStr(*(self.chunks + [Chunk(other)]))

        return NotImplemented

    def __radd__(self, other: Union["FmtStr", str]) -> "FmtStr":
        if isinstance(other, FmtStr):
            return FmtStr(*(x for x in (other.chunks + self.chunks)))
        elif isinstance(other, (bytes, str)):
            return FmtStr(*(x for x in ([Chunk(other)] + self.chunks)))

        return NotImplemented

    def __mul__(self, other: int) -> "FmtStr":
        if isinstance(other, int):
            return sum((self for _ in range(other)), FmtStr())

        return NotImplemented

    # TODO ensure empty FmtStr isn't a problem

    @property
    def shared_atts(self) -> dict[str, int | bool]:
        """Gets atts shared among all nonzero length component Chunks"""
        # TODO cache this, could get ugly for large FmtStrs
        atts = {}
        first = self.chunks[0]
        for att in sorted(first.atts):
            # TODO how to write this without the '???'?
            if all(
                fs.atts.get(att, "???") == first.atts[att]
                for fs in self.chunks
                if len(fs) > 0
            ):
                atts[att] = first.atts[att]
        return atts

    def new_with_atts_removed(self, *attributes: str) -> "FmtStr":
        """Returns a new FmtStr with the same content but some attributes removed"""

        result = FmtStr(*(Chunk(bfs.s, bfs.atts.remove(*attributes)) for bfs in self.chunks))  # type: ignore
        return result

    @no_type_check
    def __getattr__(self, att: str):
        # thanks to @aerenchyma/@jczett
        if not hasattr(self.s, att):
            raise AttributeError(f"No attribute {att!r}")

        @no_type_check
        def func_help(*args, **kwargs):
            result = getattr(self.s, att)(*args, **kwargs)
            if isinstance(result, (bytes, str)):
                return fmtstr(result, **self.shared_atts)
            elif isinstance(result, list):
                return [fmtstr(x, **self.shared_atts) for x in result]
            else:
                return result

        return func_help

    @property
    def divides(self) -> list[int]:
        """List of indices of divisions between the constituent chunks."""
        acc = [0]
        for s in self.chunks:
            acc.append(acc[-1] + len(s))
        return acc

    @property
    def s(self) -> str:
        if self._s is not None:
            return self._s
        self._s = "".join(fs.s for fs in self.chunks)
        return self._s

    def __getitem__(self, index: int | slice) -> "FmtStr":
        index = normalize_slice(len(self), index)
        counter = 0
        parts = []
        for chunk in self.chunks:
            if index.start < counter + len(chunk) and index.stop > counter:
                start = max(0, index.start - counter)
                end = min(index.stop - counter, len(chunk))
                if end - start == len(chunk):
                    parts.append(chunk)
                else:
                    s_part = chunk.s[
                        max(0, index.start - counter) : index.stop - counter
                    ]
                    parts.append(Chunk(s_part, chunk.atts))
            counter += len(chunk)
            if index.stop < counter:
                break
        return FmtStr(*parts) if parts else fmtstr("")

    def width_aware_slice(self, index: int | slice) -> "FmtStr":
        """Slice based on the number of columns it would take to display the substring."""
        if wcswidth(self.s, None) == -1:
            raise ValueError("bad values for width aware slicing")
        index = normalize_slice(self.width, index)
        counter = 0
        parts = []
        for chunk in self.chunks:
            if index.start < counter + chunk.width and index.stop > counter:
                start = max(0, index.start - counter)
                end = min(index.stop - counter, chunk.width)
                if end - start == chunk.width:
                    parts.append(chunk)
                else:
                    s_part = width_aware_slice(
                        chunk.s, max(0, index.start - counter), index.stop - counter
                    )
                    parts.append(Chunk(s_part, chunk.atts))
            counter += chunk.width
            if index.stop < counter:
                break
        return FmtStr(*parts) if parts else fmtstr("")

    def width_aware_splitlines(self, columns: int) -> Iterator["FmtStr"]:
        """Split into lines, pushing doublewidth characters at the end of a line to the next line.

        When a double-width character is pushed to the next line, a space is added to pad out the line.
        """
        if columns < 2:
            raise ValueError("Column width %s is too narrow." % columns)
        if wcswidth(self.s, None) == -1:
            raise ValueError("bad values for width aware slicing")
        return self._width_aware_splitlines(columns)

    def _width_aware_splitlines(self, columns: int) -> Iterator["FmtStr"]:
        splitter = self.chunks[0].splitter()
        chunks_of_line = []
        width_of_line = 0
        for source_chunk in self.chunks:
            splitter.reinit(source_chunk)
            while True:
                request = splitter.request(columns - width_of_line)
                if request is None:
                    break  # done with this source_chunk
                w, new_chunk = request
                chunks_of_line.append(new_chunk)
                width_of_line += w

                if width_of_line == columns:
                    yield FmtStr(*chunks_of_line)
                    del chunks_of_line[:]
                    width_of_line = 0

        if chunks_of_line:
            yield FmtStr(*chunks_of_line)

    def _getitem_normalized(self, index: int | slice) -> "FmtStr":
        """Builds the more compact fmtstrs by using fromstr( of the control sequences)"""
        index = normalize_slice(len(self), index)
        counter = 0
        output = ""
        for fs in self.chunks:
            if index.start < counter + len(fs) and index.stop > counter:
                s_part = fs.s[max(0, index.start - counter) : index.stop - counter]
                piece = Chunk(s_part, fs.atts).color_str
                output += piece
            counter += len(fs)
            if index.stop < counter:
                break
        return fmtstr(output)

    def __setitem__(self, index: int, value: Any) -> None:
        raise Exception("No!")

    def copy(self) -> "FmtStr":
        return FmtStr(*self.chunks)


def interval_overlap(a: int, b: int, x: int, y: int) -> int:
    """Returns by how much two intervals overlap

    assumed that a <= b and x <= y"""
    if b <= x or a >= y:
        return 0
    elif x <= a <= y:
        return min(b, y) - a
    elif x <= b <= y:
        return b - max(a, x)
    elif a >= x and b <= y:
        return b - a
    else:
        assert False


def width_aware_slice(s: str, start: int, end: int, replacement_char: str = " ") -> str:
    """
    >>> width_aware_slice(u'a\uff25iou', 0, 2)[1] == u' '
    True
    """
    divides = [0]
    for c in s:
        divides.append(divides[-1] + wcwidth(c))

    new_chunk_chars = []
    for char, char_start, char_end in zip(s, divides[:-1], divides[1:]):
        if char_start == start and char_end == start:
            continue  # don't use zero-width characters at the beginning of a slice
            # (combining characters combine with the chars before themselves)
        elif char_start >= start and char_end <= end:
            new_chunk_chars.append(char)
        else:
            new_chunk_chars.extend(
                replacement_char * interval_overlap(char_start, char_end, start, end)
            )

    return "".join(new_chunk_chars)


def linesplit(string: str | FmtStr, columns: int) -> list[FmtStr]:
    """Returns a list of lines, split on the last possible space of each line.

    Split spaces will be removed. Whitespaces will be normalized to one space.
    Spaces will be the color of the first whitespace character of the
    normalized whitespace.
    If a word extends beyond the line, wrap it anyway.

    >>> linesplit(fmtstr(" home    is where the heart-eating mummy is", 'blue'), 10)
    [blue('home')+blue(' ')+blue('is'), blue('where')+blue(' ')+blue('the'), blue('heart-eati'), blue('ng')+blue(' ')+blue('mummy'), blue('is')]
    """
    if not isinstance(string, FmtStr):
        string = fmtstr(string)

    string_s = string.s
    matches = list(re.finditer(r"\s+", string_s))
    spaces = [
        string[m.start() : m.end()]
        for m in matches
        if m.start() != 0 and m.end() != len(string_s)
    ]
    words = [
        string[start:end]
        for start, end in zip(
            [0] + [m.end() for m in matches],
            [m.start() for m in matches] + [len(string_s)],
        )
        if start != end
    ]

    word_to_lines = lambda word: [
        word[columns * i : columns * (i + 1)]
        for i in range((len(word) - 1) // columns + 1)
    ]

    lines = word_to_lines(words[0])
    for word, space in zip(words[1:], spaces):
        if len(lines[-1]) + len(word) < columns:
            lines[-1] += fmtstr(" ", **space.shared_atts)
            lines[-1] += word
        else:
            lines.extend(word_to_lines(word))
    return lines


def normalize_slice(length: int, index: int | slice) -> slice:
    "Fill in the Nones in a slice."
    is_int = False
    if isinstance(index, int):
        is_int = True
        index = slice(index, index + 1)
    if index.start is None:
        index = slice(0, index.stop, index.step)
    if index.stop is None:
        index = slice(index.start, length, index.step)
    if index.start < -1:  # XXX why must this be -1?
        index = slice(length - index.start, index.stop, index.step)
    if index.stop < -1:  # XXX why must this be -1?
        index = slice(index.start, length - index.stop, index.step)
    if index.step is not None:
        raise NotImplementedError("You can't use steps with slicing yet")
    if is_int:
        if index.start < 0 or index.start > length:
            raise IndexError(f"index out of bounds: {index!r} for length {length}")
    return index


def parse_args(
    args: tuple[str, ...],
    kwargs: MutableMapping[str, int | bool | str],
) -> MutableMapping[str, int | bool]:
    """Returns a kwargs dictionary by turning args into kwargs"""
    if "style" in kwargs:
        args += (cast(str, kwargs["style"]),)
        del kwargs["style"]
    for arg in args:
        if not isinstance(arg, str):
            raise ValueError(f"args must be strings: {arg!r}")
        if arg.lower() in FG_COLORS:
            if "fg" in kwargs:
                raise ValueError("fg specified twice")
            kwargs["fg"] = FG_COLORS[cast(str, arg)]
        elif arg.lower().startswith("on_") and arg[3:].lower() in BG_COLORS:
            if "bg" in kwargs:
                raise ValueError("fg specified twice")
            kwargs["bg"] = BG_COLORS[cast(str, arg[3:])]
        elif arg.lower() in STYLES:
            kwargs[arg] = True
        else:
            raise ValueError(f"couldn't process arg: {args!r}")
    for k in kwargs:
        if k not in ("fg", "bg") and k not in STYLES.keys():
            raise ValueError("Can't apply that transformation")
    if "fg" in kwargs:
        if kwargs["fg"] in FG_COLORS:
            kwargs["fg"] = FG_COLORS[cast(str, kwargs["fg"])]
        if kwargs["fg"] not in list(FG_COLORS.values()):
            raise ValueError(f"Bad fg value: {kwargs['fg']!r}")
    if "bg" in kwargs:
        if kwargs["bg"] in BG_COLORS:
            kwargs["bg"] = BG_COLORS[cast(str, kwargs["bg"])]
        if kwargs["bg"] not in list(BG_COLORS.values()):
            raise ValueError(f"Bad bg value: {kwargs['bg']!r}")
    return cast(MutableMapping[str, Union[int, bool]], kwargs)


def fmtstr(string: str | FmtStr, *args: Any, **kwargs: Any) -> FmtStr:
    """
    Convenience function for creating a FmtStr

    >>> fmtstr('asdf', 'blue', 'on_red', 'bold')
    on_red(bold(blue('asdf')))
    >>> fmtstr('blarg', fg='blue', bg='red', bold=True)
    on_red(bold(blue('blarg')))
    """
    atts = parse_args(args, kwargs)
    if isinstance(string, str):
        string = FmtStr.from_str(string)
    elif not isinstance(string, FmtStr):
        raise ValueError(
            f"Bad Args: {string!r} (of type {type(string)}), {args!r}, {kwargs!r}"
        )
    return string.copy_with_new_atts(**atts)



"""Events for keystrokes and other input events"""

import codecs
import itertools
import sys
from enum import Enum, auto
from typing import Optional, List, Union
from collections.abc import Sequence

from .termhelpers import Termmode
from .curtsieskeys import CURTSIES_NAMES as special_curtsies_names

chr_byte = lambda i: chr(i).encode("latin-1")
chr_uni = chr


CURTSIES_NAMES = {chr_byte(i): "<Ctrl-%s>" % chr(i + 0x60) for i in range(0x00, 0x1B)}
for i in range(0x00, 0x80):
    CURTSIES_NAMES[b"\x1b" + chr_byte(i)] = "<Esc+%s>" % chr(i)
for i in range(0x00, 0x1B):  # Overwrite the control keys with better labels
    CURTSIES_NAMES[b"\x1b" + chr_byte(i)] = "<Esc+Ctrl-%s>" % chr(i + 0x40)
for i in range(0x00, 0x80):
    CURTSIES_NAMES[chr_byte(i + 0x80)] = "<Meta-%s>" % chr(i)
for i in range(0x00, 0x1B):  # Overwrite the control keys with better labels
    CURTSIES_NAMES[chr_byte(i + 0x80)] = "<Meta-Ctrl-%s>" % chr(i + 0x40)

CURTSIES_NAMES.update(special_curtsies_names)

CURSES_NAMES = {
    b"\x1bOP": "KEY_F(1)",
    b"\x1bOQ": "KEY_F(2)",
    b"\x1bOR": "KEY_F(3)",
    b"\x1bOS": "KEY_F(4)",
    b"\x1b[15~": "KEY_F(5)",
    b"\x1b[17~": "KEY_F(6)",
    b"\x1b[18~": "KEY_F(7)",
    b"\x1b[19~": "KEY_F(8)",
    b"\x1b[20~": "KEY_F(9)",
    b"\x1b[21~": "KEY_F(10)",
    b"\x1b[23~": "KEY_F(11)",
    b"\x1b[24~": "KEY_F(12)",
    # see bpython #626
    b"\x1b[11~": "KEY_F(1)",
    b"\x1b[12~": "KEY_F(2)",
    b"\x1b[13~": "KEY_F(3)",
    b"\x1b[14~": "KEY_F(4)",
    b"\x1b[A": "KEY_UP",
    b"\x1b[B": "KEY_DOWN",
    b"\x1b[C": "KEY_RIGHT",
    b"\x1b[D": "KEY_LEFT",
    b"\x1b[F": "KEY_END",  # https://github.com/bpython/bpython/issues/490
    b"\x1b[H": "KEY_HOME",  # https://github.com/bpython/bpython/issues/490
    b"\x08": "KEY_BACKSPACE",
    b"\x1b[Z": "KEY_BTAB",
    # see curtsies #78 - taken from https://github.com/jquast/blessed/blob/e9ad7b85dfcbbba49010ab8c13e3a5920d81b010/blessed/keyboard.py#L409
    b"\x1b[1~": "KEY_FIND",  # find
    b"\x1b[2~": "KEY_IC",  # insert (0)
    b"\x1b[3~": "KEY_DC",  # delete (.), "Execute"
    b"\x1b[4~": "KEY_SELECT",  # select
    b"\x1b[5~": "KEY_PPAGE",  # pgup   (9)
    b"\x1b[6~": "KEY_NPAGE",  # pgdown (3)
    b"\x1b[7~": "KEY_HOME",  # home
    b"\x1b[8~": "KEY_END",  # end
    b"\x1b[OA": "KEY_UP",  # up     (8)
    b"\x1b[OB": "KEY_DOWN",  # down   (2)
    b"\x1b[OC": "KEY_RIGHT",  # right  (6)
    b"\x1b[OD": "KEY_LEFT",  # left   (4)
    b"\x1b[OF": "KEY_END",  # end    (1)
    b"\x1b[OH": "KEY_HOME",  # home   (7)
}

KEYMAP_PREFIXES = set()
for table in (CURSES_NAMES, CURTSIES_NAMES):
    for k in table:
        if k.startswith(b"\x1b"):
            for i in range(1, len(k)):
                KEYMAP_PREFIXES.add(k[:i])

MAX_KEYPRESS_SIZE = max(
    len(seq) for seq in itertools.chain(CURSES_NAMES.keys(), CURTSIES_NAMES.keys())
)


class Keynames(Enum):
    CURTSIES = auto()
    CURSES = auto()
    BYTES = auto()


class Event:
    pass


class ScheduledEvent(Event):
    """Event scheduled for a future time.

    args:
        when (float): unix time in seconds for which this event is scheduled

    Custom events that occur at a specific time in the future should
    be subclassed from ScheduledEvent."""

    def __init__(self, when: float) -> None:
        self.when = when


class WindowChangeEvent(Event):
    def __init__(self, rows: int, columns: int, cursor_dy: int | None = None) -> None:
        self.rows = rows
        self.columns = columns
        self.cursor_dy = cursor_dy

    x = width = property(lambda self: self.columns)
    y = height = property(lambda self: self.rows)

    def __repr__(self) -> str:
        return "<WindowChangeEvent (%d, %d)%s>" % (
            self.rows,
            self.columns,
            "" if self.cursor_dy is None else " cursor_dy: %d" % self.cursor_dy,
        )

    @property
    def name(self) -> str:
        return "<WindowChangeEvent>"


class SigIntEvent(Event):
    """Event signifying a SIGINT"""

    def __repr__(self) -> str:
        return "<SigInt Event>"

    @property
    def name(self) -> str:
        return repr(self)


class PasteEvent(Event):
    """Multiple keypress events combined, likely from copy/paste.

    The events attribute contains a list of keypress event strings.
    """

    def __init__(self) -> None:
        self.events: list[str] = []

    def __repr__(self) -> str:
        return "<Paste Event with data: %r>" % self.events

    @property
    def name(self) -> str:
        return repr(self)


def decodable(seq: bytes, encoding: str) -> bool:
    try:
        u = seq.decode(encoding)
    except UnicodeDecodeError:
        return False
    else:
        return True


def _key_name(seq: bytes, encoding: str, keynames: Keynames) -> str:
    if keynames == Keynames.CURSES:
        # may not be here (and still not decodable) curses names incomplete
        if seq in CURSES_NAMES:
            return CURSES_NAMES[seq]

        # Otherwise, there's no special curses name for this
        try:
            # for normal decodable text or a special curtsies sequence with bytes that can be decoded
            return seq.decode(encoding)
        except UnicodeDecodeError:
            # this sequence can't be decoded with this encoding, so we need to represent the bytes
            if len(seq) == 1:
                return "x%02X" % ord(seq)
                # TODO figure out a better thing to return here
            else:
                raise NotImplementedError(
                    "are multibyte unnameable sequences possible?"
                )
                return "bytes: " + "-".join(
                    "x%02X" % ord(seq[i : i + 1]) for i in range(len(seq))
                )
                # TODO if this isn't possible, return multiple meta keys as a paste event if paste events enabled
    elif keynames == Keynames.CURTSIES:
        if seq in CURTSIES_NAMES:
            return CURTSIES_NAMES[seq]
        # assumes that curtsies names are a subset of curses ones
        return seq.decode(encoding)
    else:
        assert keynames == Keynames.BYTES
        return seq  # type: ignore


def get_key(
    bytes_: Sequence[bytes],
    encoding: str,
    keynames: Keynames = Keynames.CURTSIES,
    full: bool = False,
) -> str | None:
    """Return key pressed from bytes_ or None

    Return a key name or None meaning it's an incomplete sequence of bytes
    (more bytes needed to determine the key pressed)

    encoding is how the bytes should be translated to unicode - it should
    match the terminal encoding.

    keynames is a string describing how keys should be named:

    * curtsies uses unicode strings like <F8>

    * curses uses unicode strings similar to those returned by
      the Python ncurses window.getkey function, like KEY_F(8),
      plus a nonstandard representation of meta keys (bytes 128-255)
      because returning the corresponding unicode code point would be
      indistinguishable from the multibyte sequence that encodes that
      character in the current encoding

    * bytes returns the original bytes from stdin (NOT unicode)

    if full, match a key even if it could be a prefix to another key
    (useful for detecting a plain escape key for instance, since
    escape is also a prefix to a bunch of char sequences for other keys)

    Events are subclasses of Event, or unicode strings

    Precondition: get_key(prefix, keynames) is None for all proper prefixes of
    bytes. This means get_key should be called on progressively larger inputs
    (for 'asdf', first on 'a', then on 'as', then on 'asd' - until a non-None
    value is returned)
    """
    if not all(isinstance(c, bytes) for c in bytes_):
        raise TypeError("get key expects bytes, got %r" % bytes_)  # expects raw bytes
    seq = b"".join(bytes_)
    if len(seq) > MAX_KEYPRESS_SIZE:
        raise ValueError("unable to decode bytes %r" % seq)

    key_known = seq in CURTSIES_NAMES or seq in CURSES_NAMES or decodable(seq, encoding)

    if full and key_known:
        return _key_name(seq, encoding, keynames)
    elif seq in KEYMAP_PREFIXES or could_be_unfinished_char(seq, encoding):
        return None  # need more input to make up a full keypress
    elif key_known:
        return _key_name(seq, encoding, keynames)
    else:
        # this will raise a unicode error (they're annoying to raise ourselves)
        seq.decode(encoding)
        assert False, "should have raised an unicode decode error"


def could_be_unfinished_char(seq: bytes, encoding: str) -> bool:
    """Whether seq bytes might create a char in encoding if more bytes were added"""
    if decodable(seq, encoding):
        return False  # any sensible encoding surely doesn't require lookahead (right?)
        # (if seq bytes encoding a character, adding another byte shouldn't also encode something)

    if codecs.getdecoder("utf8") is codecs.getdecoder(encoding):
        return could_be_unfinished_utf8(seq)
    elif codecs.getdecoder("ascii") is codecs.getdecoder(encoding):
        return False
    else:
        return True  # We don't know, it could be


def could_be_unfinished_utf8(seq: bytes) -> bool:
    # http://en.wikipedia.org/wiki/UTF-8#Description
    o = ord(seq[0:1])
    return (
        (o & 0b11100000 == 0b11000000 and len(seq) < 2)
        or (o & 0b11110000 == 0b11100000 and len(seq) < 3)
        or (o & 0b11111000 == 0b11110000 and len(seq) < 4)
        or (o & 0b11111100 == 0b11111000 and len(seq) < 5)
        or (o & 0b11111110 == 0b11111100 and len(seq) < 6)
    )


def pp_event(seq: Event | str) -> str | bytes:
    """Returns pretty representation of an Event or keypress"""

    if isinstance(seq, Event):
        return str(seq)

    # Get the original sequence back if seq is a pretty name already
    rev_curses = {v: k for k, v in CURSES_NAMES.items()}
    rev_curtsies = {v: k for k, v in CURTSIES_NAMES.items()}
    bytes_seq: bytes | None = None
    if seq in rev_curses:
        bytes_seq = rev_curses[seq]
    elif seq in rev_curtsies:
        bytes_seq = rev_curtsies[seq]

    if bytes_seq:
        pretty = curtsies_name(bytes_seq)
        if pretty != seq:
            return pretty
    return repr(seq).lstrip("u")[1:-1]


def curtsies_name(seq: bytes) -> str | bytes:
    return CURTSIES_NAMES.get(seq, seq)


def try_keys() -> None:
    print(
        "press a bunch of keys (not at the same time, but you can hit them pretty quickly)"
    )
    import os
    from .termhelpers import Cbreak

    def ask_what_they_pressed(seq: bytes, Normal: Termmode) -> None:
        print("Unidentified character sequence!")
        with Normal:
            while True:
                r = input("type 'ok' to prove you're not pounding keys ")
                if r.lower().strip() == "ok":
                    break
        while True:
            print(f"Press the key that produced {seq!r} again please")
            retry = os.read(sys.stdin.fileno(), 1000)
            if seq == retry:
                break
            print("nope, that wasn't it")
        with Normal:
            name = input("Describe in English what key you pressed: ")
            f = open("keylog.txt", "a")
            f.write(f"{seq!r} is called {name}\n")
            f.close()
            print(
                "Thanks! Please open an issue at https://github.com/bpython/curtsies/issues"
            )
            print(
                "or email thomasballinger@gmail.com. Include this terminal history or keylog.txt."
            )
            print("You can keep pressing keys")

    with Cbreak(sys.stdin) as NoCbreak:
        while True:
            try:
                chars = os.read(sys.stdin.fileno(), 1000)
                print("---")
                print(repr(chars))
                if chars in CURTSIES_NAMES:
                    print(CURTSIES_NAMES[chars])
                elif len(chars) == 1:
                    print("literal")
                else:
                    print("unknown!!!")
                    ask_what_they_pressed(chars, NoCbreak)
            except OSError:
                pass


if __name__ == "__main__":
    try_keys()



"""Constants for terminal formatting"""

from collections.abc import Mapping

colors = "black", "red", "green", "yellow", "blue", "magenta", "cyan", "gray"
FG_COLORS: Mapping[str, int] = dict(zip(colors, range(30, 38)))
BG_COLORS: Mapping[str, int] = dict(zip(colors, range(40, 48)))
STYLES: Mapping[str, int] = dict(
    zip(("bold", "dark", "italic", "underline", "blink", "invert"), (1, 2, 3, 4, 5, 7))
)
FG_NUMBER_TO_COLOR: Mapping[int, str] = dict(zip(FG_COLORS.values(), FG_COLORS.keys()))
BG_NUMBER_TO_COLOR: Mapping[int, str] = dict(zip(BG_COLORS.values(), BG_COLORS.keys()))
NUMBER_TO_STYLE = dict(zip(STYLES.values(), STYLES.keys()))
RESET_ALL = 0
RESET_FG = 39
RESET_BG = 49


def seq(num: int) -> str:
    return f"[{num}m"



import tty
import termios
import fcntl
import os

from typing import IO, ContextManager, Type, List, Union, Optional
from types import TracebackType

_Attr = list[Union[int, list[Union[bytes, int]]]]


class Nonblocking(ContextManager):
    """
    A context manager for making an input stream nonblocking.
    """

    def __init__(self, stream: IO) -> None:
        self.stream = stream
        self.fd = self.stream.fileno()

    def __enter__(self) -> None:
        self.orig_fl = fcntl.fcntl(self.fd, fcntl.F_GETFL)
        fcntl.fcntl(self.fd, fcntl.F_SETFL, self.orig_fl | os.O_NONBLOCK)

    def __exit__(
        self,
        type: type[BaseException] | None = None,
        value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        fcntl.fcntl(self.fd, fcntl.F_SETFL, self.orig_fl)


class Termmode(ContextManager):
    def __init__(self, stream: IO, attrs: _Attr) -> None:
        self.stream = stream
        self.attrs = attrs

    def __enter__(self) -> None:
        self.original_stty = termios.tcgetattr(self.stream)
        termios.tcsetattr(self.stream, termios.TCSANOW, self.attrs)

    def __exit__(
        self,
        type: type[BaseException] | None = None,
        value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        termios.tcsetattr(self.stream, termios.TCSANOW, self.original_stty)


class Cbreak(ContextManager[Termmode]):
    def __init__(self, stream: IO) -> None:
        self.stream = stream

    def __enter__(self) -> Termmode:
        self.original_stty = termios.tcgetattr(self.stream)
        tty.setcbreak(self.stream, termios.TCSANOW)
        return Termmode(self.stream, self.original_stty)

    def __exit__(
        self,
        type: type[BaseException] | None = None,
        value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        termios.tcsetattr(self.stream, termios.TCSANOW, self.original_stty)



r"""
Parses ascii escape sequences into marked up strings

>>> r = peel_off_esc_code('Amore')
>>> r == ('', {'csi': '\x1b', 'command': 'A', 'seq': '\x1bA'}, 'more')
True
>>> r = peel_off_esc_code('[2Astuff')
>>> r == ('', {'csi': '\x1b[', 'seq': '\x1b[2A', 'intermed': '', 'private': '', 'command': 'A', 'numbers': [2]}, 'stuff')
True
"""

from typing import (
    List,
    Union,
    Tuple,
    cast,
    Dict,
    Any,
    Optional,
)
from collections.abc import Mapping
from re import Match

import re

from .termformatconstants import (
    FG_NUMBER_TO_COLOR,
    BG_NUMBER_TO_COLOR,
    NUMBER_TO_STYLE,
    RESET_ALL,
    RESET_FG,
    RESET_BG,
    STYLES,
)


Token = dict[str, Union[str, list[int]]]


def remove_ansi(s: str) -> str:
    return re.sub(r"(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]", "", s)


def parse(s: str) -> list[str | dict[str, str | bool | None]]:
    r"""
    Returns a list of strings or format dictionaries to describe the strings.

    May raise a ValueError if it can't be parsed.

    >>> parse(">>> []")
    ['>>> []']
    >>> parse("\x1b[33m[\x1b[39m\x1b[33m]\x1b[39m\x1b[33m[\x1b[39m\x1b[33m]\x1b[39m\x1b[33m[\x1b[39m\x1b[33m]\x1b[39m\x1b[33m[\x1b[39m")
    [{'fg': 'yellow'}, '[', {'fg': None}, {'fg': 'yellow'}, ']', {'fg': None}, {'fg': 'yellow'}, '[', {'fg': None}, {'fg': 'yellow'}, ']', {'fg': None}, {'fg': 'yellow'}, '[', {'fg': None}, {'fg': 'yellow'}, ']', {'fg': None}, {'fg': 'yellow'}, '[', {'fg': None}]
    """
    stuff: list[str | dict[str, str | bool | None]] = []
    rest = s
    while True:
        front, token, rest = peel_off_esc_code(rest)
        if front:
            stuff.append(front)
        if token:
            try:
                tok = token_type(token)
                if tok:
                    stuff.extend(tok)
            except ValueError:
                raise ValueError(
                    "Can't parse escape sequence: %r %r %r %r"
                    % (s, repr(front), token, repr(rest))
                )
        if not rest:
            break
    return stuff


def peel_off_esc_code(s: str) -> tuple[str, Token | None, str]:
    r"""Returns processed text, the next token, and unprocessed text

    >>> front, d, rest = peel_off_esc_code('some[2Astuff')
    >>> front, rest
    ('some', 'stuff')
    >>> d == {'numbers': [2], 'command': 'A', 'intermed': '', 'private': '', 'csi': '\x1b[', 'seq': '\x1b[2A'}
    True
    """
    # fmt: off
    p = r"""(?P<front>.*?)
            (?P<seq>
                (?P<csi>
                    (?:[]\[)
                    |
                    ["""+'\x9b' + r"""])
                (?P<private>)
                (?P<numbers>
                    (?:\d+;)*
                    (?:\d+)?)
                (?P<intermed>""" + '[\x20-\x2f]*)' + r"""
                (?P<command>""" + '[\x40-\x7e]))' + r"""
            (?P<rest>.*)"""
    # fmt: on
    m1 = re.match(p, s, re.VERBOSE)  # multibyte esc seq
    m2 = re.match(
        "(?P<front>.*?)(?P<seq>(?P<csi>)(?P<command>[\x40-\x5f]))(?P<rest>.*)", s
    )  # 2 byte escape sequence
    m = None  # Optional[Match[str]]
    if m1 and m2:
        m = m1 if len(m1.groupdict()["front"]) <= len(m2.groupdict()["front"]) else m2
        # choose the match which has less processed text in order to get the
        # first escape sequence
    elif m1:
        m = m1
    elif m2:
        m = m2
    else:
        m = None

    if m:
        d: dict[str, Any] = m.groupdict()
        del d["front"]
        del d["rest"]
        if "numbers" in d and all(d["numbers"].split(";")):
            d["numbers"] = [int(x) for x in d["numbers"].split(";")]

        return m.groupdict()["front"], cast(Token, d), m.groupdict()["rest"]
    else:
        return s, None, ""


def token_type(info: Token) -> list[dict[str, str | bool | None]] | None:
    if info["command"] == "m":
        # The default action for ESC[m is to act like ESC[0m
        # Ref: https://en.wikipedia.org/wiki/ANSI_escape_code#CSI_codes
        values = cast(list[int], info["numbers"]) if len(info["numbers"]) else [0]
        tokens: list[dict[str, str | bool | None]] = []
        for value in values:
            if value in FG_NUMBER_TO_COLOR:
                tokens.append({"fg": FG_NUMBER_TO_COLOR[value]})
            if value in BG_NUMBER_TO_COLOR:
                tokens.append({"bg": BG_NUMBER_TO_COLOR[value]})
            if value in NUMBER_TO_STYLE:
                tokens.append({NUMBER_TO_STYLE[value]: True})
            if value == RESET_ALL:
                tokens.append(
                    dict({k: None for k in STYLES}, **{"fg": None, "bg": None})
                )
            if value == RESET_FG:
                tokens.append({"fg": None})
            if value == RESET_BG:
                tokens.append({"bg": None})

        if tokens:
            return tokens
        else:
            raise ValueError("Can't parse escape seq %r" % info)
    elif info["command"] == "H":  # fix for bpython #76
        return [{}]
    return None


if __name__ == "__main__":
    import doctest

    doctest.testmod()
    # print(peel_off_esc_code('[2Astuff'))
    # print(peel_off_esc_code('Amore'))
    print(repr(parse("[31mstuff is the best[32myay")))



"""All the key sequences"""

# If you add a binding, add something about your setup
# if you can figure out why it's different

# Special names are for multi-character keys, or key names
# that would be hard to write in a config file

# TODO add PAD keys hack as in bpython.cli

# fmt: off
CURTSIES_NAMES = {
    b' ':          '<SPACE>',
    b'\x1b ':      '<Esc+SPACE>',
    b'\t':         '<TAB>',
    b'\x1b[Z':     '<Shift-TAB>',
    b'\x1b[A':     '<UP>',
    b'\x1b[B':     '<DOWN>',
    b'\x1b[C':     '<RIGHT>',
    b'\x1b[D':     '<LEFT>',
    b'\x1bOA':     '<UP>',         # in issue 92 its shown these should be normal arrows,
    b'\x1bOB':     '<DOWN>',       # not ctrl-arrows as we previously had them.
    b'\x1bOC':     '<RIGHT>',
    b'\x1bOD':     '<LEFT>',

    b'\x1b[1;5A':  '<Ctrl-UP>',
    b'\x1b[1;5B':  '<Ctrl-DOWN>',
    b'\x1b[1;5C':  '<Ctrl-RIGHT>', # reported by myint
    b'\x1b[1;5D':  '<Ctrl-LEFT>',  # reported by myint

    b'\x1b[5A':    '<Ctrl-UP>',    # not sure about these, someone wanted them for bpython
    b'\x1b[5B':    '<Ctrl-DOWN>',
    b'\x1b[5C':    '<Ctrl-RIGHT>',
    b'\x1b[5D':    '<Ctrl-LEFT>',

    b'\x1b[1;9A':  '<Esc+UP>',
    b'\x1b[1;9B':  '<Esc+DOWN>',
    b'\x1b[1;9C':  '<Esc+RIGHT>',
    b'\x1b[1;9D':  '<Esc+LEFT>',

    b'\x1b[1;10A': '<Esc+Shift-UP>',
    b'\x1b[1;10B': '<Esc+Shift-DOWN>',
    b'\x1b[1;10C': '<Esc+Shift-RIGHT>',
    b'\x1b[1;10D': '<Esc+Shift-LEFT>',

    b'\x1bOP':     '<F1>',
    b'\x1bOQ':     '<F2>',
    b'\x1bOR':     '<F3>',
    b'\x1bOS':     '<F4>',

    # see bpython #626
    b'\x1b[11~':   '<F1>',
    b'\x1b[12~':   '<F2>',
    b'\x1b[13~':   '<F3>',
    b'\x1b[14~':   '<F4>',

    b'\x1b[15~':   '<F5>',
    b'\x1b[17~':   '<F6>',
    b'\x1b[18~':   '<F7>',
    b'\x1b[19~':   '<F8>',
    b'\x1b[20~':   '<F9>',
    b'\x1b[21~':   '<F10>',
    b'\x1b[23~':   '<F11>',
    b'\x1b[24~':   '<F12>',
    b'\x00':       '<Ctrl-SPACE>',
    b'\x1c':       '<Ctrl-\\>',
    b'\x1d':       '<Ctrl-]>',
    b'\x1e':       '<Ctrl-6>',
    b'\x1f':       '<Ctrl-/>',
    b'\x7f':       '<BACKSPACE>',    # for some folks this is ctrl-backspace apparently
    b'\x1b\x7f':   '<Esc+BACKSPACE>',
    b'\xff':       '<Meta-BACKSPACE>',
    b'\x1b\x1b[A': '<Esc+UP>',    # uncertain about these four
    b'\x1b\x1b[B': '<Esc+DOWN>',
    b'\x1b\x1b[C': '<Esc+RIGHT>',
    b'\x1b\x1b[D': '<Esc+LEFT>',
    b'\x1b':       '<ESC>',
    b'\x1b[1~':    '<HOME>',
    b'\x1b[4~':    '<END>',
    b'\x1b\x1b[5~':'<Esc+PAGEUP>',
    b'\x1b\x1b[6~':'<Esc+PAGEDOWN>',

    b'\x1b[H':     '<HOME>',    # reported by amorozov in bpython #490
    b'\x1b[F':     '<END>',     # reported by amorozov in bpython #490

    b'\x1bOH':     '<HOME>',    # reported by mixmastamyk in curtsies #78
    b'\x1bOF':     '<END>',     # reported by mixmastamyk in curtsies #78

    # not fixing for back compat.
    # (b"\x1b[1~": u'<FIND>',       # find

    b"\x1b[2~": '<INSERT>',       # insert (0)
    b"\x1b[3~": '<DELETE>',       # delete (.), "Execute"
    b"\x1b[3;5~": '<Ctrl-DELETE>',

    # st (simple terminal) see issue #169
    b"\x1b[4h": '<INSERT>',
    b"\x1b[P": '<DELETE>',

    # not fixing for back compat.
    # (b"\x1b[4~": u'<SELECT>',       # select

    b"\x1b[5~": '<PAGEUP>',       # pgup   (9)
    b"\x1b[6~": '<PAGEDOWN>',     # pgdown (3)
    b"\x1b[7~": '<HOME>',         # home
    b"\x1b[8~": '<END>',          # end
    b"\x1b[OA": '<UP>',           # up     (8)
    b"\x1b[OB": '<DOWN>',         # down   (2)
    b"\x1b[OC": '<RIGHT>',        # right  (6)
    b"\x1b[OD": '<LEFT>',         # left   (4)
    b"\x1b[OF": '<END>',          # end    (1)
    b"\x1b[OH": '<HOME>',         # home   (7)

    # reported by cool-RR
    b"\x1b[[A": '<F1>',
    b"\x1b[[B": '<F2>',
    b"\x1b[[C": '<F3>',
    b"\x1b[[D": '<F4>',
    b"\x1b[[E": '<F5>',
    # cool-RR says the rest were good: see issue #99

    # reported by alethiophile see issue #119
    b"\x1b[1;3C": '<Meta-RIGHT>',      # alt-right
    b"\x1b[1;3B": '<Meta-DOWN>',       # alt-down
    b"\x1b[1;3D": '<Meta-LEFT>',       # alt-left
    b"\x1b[1;3A": '<Meta-UP>',         # alt-up
    b"\x1b[5;3~": '<Meta-PAGEUP>',     # alt-pageup
    b"\x1b[6;3~": '<Meta-PAGEDOWN>',   # alt-pagedown
    b"\x1b[1;3H": '<Meta-HOME>',       # alt-home
    b"\x1b[1;3F": '<Meta-END>',        # alt-end
    b"\x1b[1;2C": '<Shift-RIGHT>',
    b"\x1b[1;2B": '<Shift-RIGHT>',
    b"\x1b[1;2D": '<Shift-RIGHT>',
    b"\x1b[1;2A": '<Shift-RIGHT>',
    b"\x1b[3;2~": '<Shift-DELETE>',
    b"\x1b[5;2~": '<Shift-PAGEUP>',
    b"\x1b[6;2~": '<Shift-PAGEDOWN>',
    b"\x1b[1;2H": '<Shift-HOME>',
    b"\x1b[1;2F": '<Shift-END>',
    # end of keys reported by alethiophile
}
# fmt: on



"""
Format String 2D array
2d array for compositing term-formated strings


-autoexpanding vertically
-interesting get_item behavior (renders fmtstrs)
-caching behavior eventually


>>> a = FSArray(10, 14)
>>> a.shape
(10, 14)
>>> a[1] = 'i'
>>> a[3:4, :] = ['i' * 14]
>>> a[16:17, :] = ['j' * 14]
>>> a.shape, a[16, 0]
((17, 14), ['j'])
>>> a[200, 1] = ['i']
>>> a[200, 1]
['i']
"""

import itertools
import sys
import logging

from .formatstring import fmtstr
from .formatstring import normalize_slice
from .formatstring import FmtStr

from typing import (
    Any,
    Optional,
    Union,
    List,
    overload,
    Tuple,
    cast,
    no_type_check,
)
from collections.abc import Sequence

logger = logging.getLogger(__name__)

# TODO check that strings used in arrays don't have tabs or spaces in them!


def slicesize(s: slice) -> int:
    return int((s.stop - s.start) / (s.step if s.step else 1))


class FSArray(Sequence):
    """A 2D array of colored text.

    Internally represented by a list of FmtStrs of identical size."""

    # TODO add constructor that takes fmtstrs instead of dims
    def __init__(
        self, num_rows: int, num_columns: int, *args: Any, **kwargs: Any
    ) -> None:
        self.saved_args, self.saved_kwargs = args, kwargs
        self.rows: list[FmtStr] = [fmtstr("", *args, **kwargs) for _ in range(num_rows)]
        self.num_columns = num_columns

    @overload
    def __getitem__(self, slicetuple: int) -> FmtStr:
        pass

    @overload
    def __getitem__(self, slicetuple: slice) -> list[FmtStr]:
        pass

    @overload
    def __getitem__(self, slicetuple: tuple[slice | int, slice | int]) -> list[FmtStr]:
        pass

    def __getitem__(
        self, slicetuple: int | slice | tuple[int | slice, int | slice]
    ) -> FmtStr | list[FmtStr]:
        if isinstance(slicetuple, int):
            if slicetuple < 0:
                slicetuple = len(self.rows) - slicetuple
            if slicetuple < 0 or slicetuple >= len(self.rows):
                raise IndexError("out of bounds")
            return self.rows[slicetuple]
        if isinstance(slicetuple, slice):
            rowslice = normalize_slice(len(self.rows), slicetuple)
            return self.rows[rowslice]
        row_slice_or_int, col_slice_or_int = slicetuple
        rowslice = normalize_slice(len(self.rows), row_slice_or_int)
        colslice = normalize_slice(self.num_columns, col_slice_or_int)
        # TODO clean up slices
        return [fs[colslice] for fs in self.rows[rowslice]]

    def __len__(self) -> int:
        return len(self.rows)

    @property
    def shape(self) -> tuple[int, int]:
        """Tuple of (len(rows, len(num_columns)) numpy-style shape"""
        return len(self.rows), self.num_columns

    @property
    def height(self) -> int:
        """The number of rows"""
        return len(self.rows)

    @property
    def width(self) -> int:
        """The number of columns"""
        return self.num_columns

    # TODO rework this next major version bump
    @no_type_check
    def __setitem__(self, slicetuple, value):
        """Place a FSArray in a FSArray"""
        logger.debug("slice: %r", slicetuple)
        if isinstance(slicetuple, slice):
            rowslice, colslice = slicetuple, slice(None)
            if isinstance(value, str):
                raise ValueError(
                    "if slice is 2D, value must be 2D as in of list type []"
                )
        elif isinstance(slicetuple, int):
            normalize_slice(self.height, slicetuple)
            self.rows[slicetuple] = value
            return
        else:
            rowslice, colslice = slicetuple

        # temp shim to allow numpy arrays as values
        if value.__class__.__name__ == "ndarray":
            value = [fmtstr("".join(line)) for line in value]

        rowslice = normalize_slice(sys.maxsize, rowslice)
        additional_rows = max(0, rowslice.stop - len(self.rows))
        self.rows.extend(
            [
                fmtstr("", *self.saved_args, **self.saved_kwargs)
                for _ in range(additional_rows)
            ]
        )
        logger.debug("num columns: %r", self.num_columns)
        logger.debug("colslice: %r", colslice)
        colslice = normalize_slice(self.num_columns, colslice)
        if slicesize(colslice) == 0 or slicesize(rowslice) == 0:
            return
        if slicesize(colslice) > 1 and isinstance(value, str):
            raise ValueError(
                """You cannot replace a multi column slice with a 
                string please use a list [] with strings for the 
                contents of each row"""
            )
        if slicesize(rowslice) != len(value):
            area = slicesize(rowslice) * slicesize(colslice)
            val_len = sum(len(i) for i in value)
            grid_value = [fmtstr(" ", bg="cyan") * slicesize(colslice)] * slicesize(
                rowslice
            )
            grid_fsarray = (
                self.rows[: rowslice.start]
                + [
                    fs.setslice_with_length(
                        colslice.start, colslice.stop, v, self.num_columns
                    )
                    for fs, v in zip(self.rows[rowslice], grid_value)
                ]
                + self.rows[rowslice.stop :]
            )
            msg = "You are trying to fit this value {} into the region {}: {}".format(
                fmtstr("".join(value), bg="cyan"),
                fmtstr("").join(grid_value),
                "\n ".join(grid_fsarray[x] for x in range(len(self.rows))),
            )
            raise ValueError(
                """Error you are trying to replace a region of {} rows by {}
                columns for and area of {} with a value of len {}. The value
                used to replace the region must equal the area of the region
                replace.
                {}""".format(
                    rowslice.stop - rowslice.start,
                    colslice.stop - colslice.start,
                    area,
                    val_len,
                    msg,
                )
            )
        self.rows = (
            self.rows[: rowslice.start]
            + [
                fs.setslice_with_length(
                    colslice.start, colslice.stop, v, self.num_columns
                )
                for fs, v in zip(self.rows[rowslice], value)
            ]
            + self.rows[rowslice.stop :]
        )

    def dumb_display(self) -> None:
        """Prints each row followed by a newline without regard for the terminal window size"""
        for line in self.rows:
            print(line)

    @classmethod
    def diff(cls, a: "FSArray", b: "FSArray", ignore_formatting: bool = False) -> str:
        """Returns two FSArrays with differences underlined"""

        def underline(x: str) -> str:
            return f"\x1b[4m{x}\x1b[0m"

        def blink(x: str) -> str:
            return f"\x1b[5m{x}\x1b[0m"

        a_rows = []
        b_rows = []
        max_width = max(len(row) for row in itertools.chain(a, b))
        a_lengths = []
        b_lengths = []
        for a_row, b_row in zip(a, b):
            a_lengths.append(len(a_row))
            b_lengths.append(len(b_row))
            extra_a = "`" * (max_width - len(a_row))
            extra_b = "`" * (max_width - len(b_row))
            a_line = ""
            b_line = ""
            for a_char, b_char in zip(a_row + extra_a, b_row + extra_b):
                if ignore_formatting:
                    a_char_for_eval = a_char.s if isinstance(a_char, FmtStr) else a_char
                    b_char_for_eval = b_char.s if isinstance(b_char, FmtStr) else b_char
                else:
                    a_char_for_eval = a_char
                    b_char_for_eval = b_char
                if a_char_for_eval == b_char_for_eval:
                    a_line += str(a_char)
                    b_line += str(b_char)
                else:
                    a_line += underline(blink(str(a_char)))
                    b_line += underline(blink(str(b_char)))
            a_rows.append(a_line)
            b_rows.append(b_line)
        return "\n".join(
            f"{a_line} {a_len:3d} | {b_len:3d} {b_line}"
            for a_line, b_line, a_len, b_len in zip(
                a_rows, b_rows, a_lengths, b_lengths
            )
        )


def fsarray(
    strings: Sequence[FmtStr | str],
    width: int | None = None,
    *args: Any,
    **kwargs: Any,
) -> FSArray:
    """fsarray(list_of_FmtStrs_or_strings, width=None) -> FSArray

    Returns a new FSArray of width of the maximum size of the provided
    strings, or width provided, and height of the number of strings provided.
    If a width is provided, raises a ValueError if any of the strings
    are of length greater than this width"""

    strings = list(strings)
    if width is not None:
        if strings and any(len(s) > width for s in strings):
            raise ValueError(f"Those strings won't fit for width {width}")
    else:
        width = max(len(s) for s in strings) if strings else 0
    arr = FSArray(len(strings), width, *args, **kwargs)
    rows = [
        fs.setslice_with_length(0, len(s), s, width)
        for fs, s in zip(
            arr.rows,
            (
                s if isinstance(s, FmtStr) else fmtstr(s, *args, **kwargs)
                for s in strings
            ),
        )
    ]
    arr.rows = rows
    return arr


def simple_format(x: FSArray | Sequence[FmtStr]) -> str:
    return "\n".join(str(l) for l in x)


def assertFSArraysEqual(a: FSArray, b: FSArray) -> None:
    assert isinstance(a, FSArray)
    assert isinstance(b, FSArray)
    assert (
        a.width == b.width and a.height == b.height
    ), f"fsarray dimensions do not match: {a.shape} {b.shape}"
    for i, (a_row, b_row) in enumerate(zip(a, b)):
        assert a_row == b_row, "FSArrays differ first on line {}:\n{}".format(
            i, FSArray.diff(a, b)
        )


def assertFSArraysEqualIgnoringFormatting(a: FSArray, b: FSArray) -> None:
    """Also accepts arrays of strings"""
    assert len(a) == len(b), "fsarray heights do not match: {} {} \n{} \n{}".format(
        len(a),
        len(b),
        simple_format(a),
        simple_format(b),
    )
    for i, (a_row, b_row) in enumerate(zip(a, b)):
        a_row = a_row.s if isinstance(a_row, FmtStr) else a_row
        b_row = b_row.s if isinstance(b_row, FmtStr) else b_row
        assert a_row == b_row, "FSArrays differ first on line {}:\n{}".format(
            i,
            FSArray.diff(a, b, ignore_formatting=True),
        )


if __name__ == "__main__":
    a = FSArray(3, 14, bg="blue")
    a[0:2, 5:11] = cast(
        tuple[FmtStr, ...],
        (fmtstr("hey", "on_blue") + " " + fmtstr("yo", "on_red"), fmtstr("qwe qw")),
    )
    a.dumb_display()

    a = fsarray(["hey", "there"], bg="cyan")
    a.dumb_display()
    print(FSArray.diff(a, fsarray(["hey", "there "]), ignore_formatting=True))
